/*
 *  $Id$
 */
 
/*
 *  sobexsrv (c) Collin R. Mulliner <collin-sobexsrv(AT)betaversion.net>
 *               web: http://www.mulliner.org/bluetooth/
 *
 *  license: GPL
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <getopt.h>
#include <bluetooth/bluetooth.h>
#include <openobex/obex.h>
#include <fcntl.h>
#include <pwd.h>

#include "obexsrv.h"
#include "util.h"
#include "sdp.h"
#include "hci.h"

server_context *context;

void debug4(char *m1, void *m2, void *m3, void *m4)
{
	if (!context->daemon) fprintf(stderr, m1, m2, m3, m4);
}

void debug(char *m1)
{
	debug4(m1, NULL, NULL, NULL);
}

void debug2(char *m1, void *m2)
{
	debug4(m1, m2, NULL, NULL);
}

void debug3(char *m1, void *m2, void *m3)
{
	debug4(m1, m2, m3, NULL);
}

int handler(char *command)
{
	#define BUFSIZE 4096
	char *buf = NULL;
	char local[20];
	char remote[20];
	int handler_res;
	int pid;
	int rd[2];
	int wr[2];
	int i;
	

	if ((buf = malloc(BUFSIZE)) == NULL) {
		debug("handler(): can't alloc memory\n");
		exit(-1);
	}
	
	ba2str(&context->local_address, local);
	ba2str(&context->remote_address, remote);
	
	snprintf(buf, BUFSIZE, "command: %s\nlocal: %s\nremote: %s\n", command, local,
		remote);
	if (context->filename != NULL) {
		snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "filename: %s\n",
			context->filename);
	}
	if (context->type != NULL) {
		snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "type: %s\n", context->type);
	}
	if (context->description != NULL) {
		snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "description: %s\n",
			context->description);
	}
	if (context->path != NULL) {
		snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "path: %s\n",
			context->path);
	}
	if (context->target != NULL) {
		if ((unsigned char)context->target[0] > 'z') {
			snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "target: ");
			for (i = 0; i < context->target_length; i++) {
				snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "%0.2X",
					(unsigned char)context->target[i]);
			}
			snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "\n");
		}
		else {
			snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "target: %s\n",
				context->target);
		}
	}
	snprintf(buf+strlen(buf), BUFSIZE-strlen(buf), "length: %u\ncount: %u\n"\
		"connectionid: %u\ndatatype: %d\n\n", context->data_length, context->count,
		context->connection_id, context->data_type);
		
	if (pipe(rd) != 0 || pipe(wr) != 0) {
		debug("handler: can't create pipes\n");
		exit(-1);
	}
	
	/* child process */
	if ((pid = fork()) == 0) {
		if (setgid(context->user->pw_gid) < 0) {
			debug("can't drop privileges!\n");
			exit(-1);
		}
		if (setuid(context->user->pw_uid)< 0) {
			debug("can't drop privileges!\n");
			exit(-1);
		}
		debug3("dropping to uid=%d and gid=%d\n", context->user->pw_uid, context->user->pw_gid);
		close(rd[0]);
		close(wr[1]);
		close(0);
		close(1);
		dup(wr[0]);
		dup(rd[1]);
		if (execl((const char*)context->handler_script, (const char*)
				context->handler_script, NULL) == -1) {
			debug("handler: execl failed\n");
			exit(-1);
		}
	}
	
	/* mother process */	
	close(rd[1]);
	close(wr[0]);
	
	debug2("handler: header buf size %d\n", strlen(buf));
	if (nwrite(wr[1], buf, strlen(buf)) <= 0) {
		debug("handler: write failed\n");
		exit(-1);
	}
	free(buf);
	buf = NULL;
	
	if (strcmp(command, "put") == 0) {
		if (context->data_type == 1) {
			if (context->data == NULL) {
				debug("handler: put with no body data\n");
				exit(-1);
			}
			if (nwrite(wr[1], context->data, context->data_length) <= 0) {
				debug("handler: put body write failed\n");
				exit(-1);
			}
		}
	}
	else if (strcmp(command, "get") == 0) {
		int i = 0;
		char *p;
		buf = malloc(1024);
		bzero(buf, 1024);
		do {
			if (cread(rd[0], &buf[i], 1) <= 0) {
				debug("handler: get read failed\n");
				//exit(-1);
				break;
			}
			if (i > 2) {
				if (buf[i-1] == '\n' && buf[i] == '\n') break;
			}
			i++;
		} while (i < 1023);
		if ((p = strstr(buf, "length:")) != NULL) {
			context->data_length = strtol(p+7, NULL, 10);
			if (context->data_length < 0) {
				debug2("handler: get length error (%d)\n", context->data_length);
				exit(-1);
			}
			debug2("handler: get length %d\n", context->data_length);
			
			if (context->data_length > 0) {
				if ((context->data = malloc(context->data_length)) == NULL) {
					debug("handler: get data malloc failed\n");
					exit(-1);
				}
				if ((cread(rd[0], context->data, context->data_length)) < context->data_length) {
					debug("handler: get data read failed\n");
					exit(-1);
				}
			}
		}
		else {
			debug("handler: get result parse error\n");
			exit(-1);
		}
	}

	/* close descriptors and wait for child */
	close(wr[1]);
	close(rd[0]);
	do {
		waitpid(pid, &handler_res, WNOHANG);
	} while (!WIFEXITED(handler_res));
	
	debug2("handler: returned with exit code %d\n", WEXITSTATUS(handler_res));
	return(WEXITSTATUS(handler_res));
}

void context_clear(int all)
{
	if (context->filename) {
		free(context->filename);
		context->filename = NULL;
	}
	if (context->type) {
		free(context->type);
		context->type = NULL;
	}
	if (context->description) {
		free(context->description);
		context->description = NULL;
	}
	context->data_length = 0;
	context->data = 0;
	context->data_type = 0;
	context->count = 0;
	
	if (all) {
		if (context->target) {
			free(context->target);
			context->target = NULL;
			context->target_length = 0;
		}
		if (context->path) {
			free(context->path);
			context->path = NULL;
		}
		context->connection_id = 0;
		bzero(&context->remote_address, sizeof(context->remote_address));
	}
}

int cmd_connect(obex_t *handle, obex_object_t *oo)
{
	obex_headerdata_t hv;
	uint8_t hi;
	int hlen;
	int i;
	

	debug("CONNECT start\n");

	while (OBEX_ObjectGetNextHeader(handle, oo, &hi, &hv, &hlen))	{
		switch (hi) {
		case OBEX_HDR_CONNECTION:
			context->connection_id = hv.bq4;
			
			debug2("CONNECT connection id: %d\n", hv.bq4);
			break;
		case OBEX_HDR_TARGET:
			if ((context->target = malloc(hlen+1))) {
				bzero(context->target, hlen+1);
				memcpy(context->target, hv.bs, hlen);
				context->target_length = hlen;
				
				debug("CONNECT target ");
				if ((unsigned char)context->target[0] > 'z') {
					for (i = 0; i < hlen; i++) {
						debug2("%0.2X", hv.bs[i]);
					}
					debug("\n");
				}
				else {
					debug2("%s\n", context->target);
				}
			}
			else {
				debug("CONNECT target malloc error\n");
			}
			break;
		default:
			debug2("CONNECT skipped header %02x\n", hi);
			break;
		}
	}

	switch (handler("connect")) {
	case -1:
		debug("CONNECT handler error\n");
		exit(-1);
		break;
	case 1:
		debug("CONNECT ok\n");
		OBEX_ObjectSetRsp(oo, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
		break;
	default:
		debug("CONNECT deny (fix me!)\n");
		// FIX ME
		OBEX_ObjectSetRsp(oo, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
		break;
	}

	debug("CONNECT end\n");

	return(1);
}

int cmd_disconnect(obex_t *handle, obex_object_t *oo)
{
	debug("DISCONNECT start\n");

	OBEX_ObjectSetRsp(oo, OBEX_RSP_SUCCESS, OBEX_RSP_SUCCESS);
	
	debug("DISCONNECT end\n");
	
	return(1);
}

int cmd_setpath(obex_t *handle, obex_object_t *oo)
{
	obex_headerdata_t hv;
	uint8_t hi;
	int hlen;
	int mkdir = 1;
	int ret;

	
	debug("SETPATH start\n");

	while (OBEX_ObjectGetNextHeader(handle, oo, &hi, &hv, &hlen))	{
		switch (hi) {
		case OBEX_HDR_NAME:
			// name header present so we don't want to create the current dir.
			mkdir = 0;
			if (hlen > 0) {
				if ((context->path = malloc(hlen))) {
					bzero(context->path, hlen);
					OBEX_UnicodeToChar(context->path, hv.bs, hlen);
					debug2("SETPATH path %s\n", context->path);
				}
			}
			else {
				debug("SETPATH path NOPATH\n");
			}
			break;
		default:
			debug2("SETPATH skipped header %02x\n", hi);
			break;
		}
	}

	ret = handler((mkdir) ? "mkdir" : "setpath");
	
	switch (ret) {
	case -1:
		debug("SETPATH handler error\n");
		exit(-1);
		break;
	case 1:
		if (mkdir) debug2("SETPATH/MKDIR %s ok\n", context->path);
		else debug("SETPATH ok\n");
		OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
		//hv.bq4 = context->connection_id;
		//OBEX_ObjectAddHeader(handle, oo, OBEX_HDR_CONNECTION, hv, sizeof(uint32_t), 0);
		break;
	default:
		if (mkdir) debug2("SETPATH/MKDIR %s failed\n", context->path);
		else debug("SETPATH deny (fix me!)\n");
		// FIX ME
		OBEX_ObjectSetRsp(oo, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
		break;
	}
	
	debug("SETPATH end\n");

	context_clear(0);
	return(1);
}

int cmd_get(obex_t *handle, obex_object_t *oo)
{
	obex_headerdata_t hv;
	uint8_t hi;
	int hlen;

	
	debug("GET start\n");

	while (OBEX_ObjectGetNextHeader(handle, oo, &hi, &hv, &hlen))	{
		switch (hi) {
		case OBEX_HDR_NAME:
			if ((context->filename = malloc(hlen))) {
				OBEX_UnicodeToChar(context->filename, hv.bs, hlen);
				
				debug2("GET filename: %s\n", context->filename);
			}
			else {
				// malloc error
				goto err;
			}
			break;
		case OBEX_HDR_TYPE:
		if ((context->type = malloc(hlen))) {
				strncpy(context->type, hv.bs, hlen);
				
				debug2("GET type: %s\n", context->type);
			}
			else {
				// malloc error
				goto err;
			}
			break;
		case OBEX_HDR_CONNECTION:
			debug2("GET connection %d\n", hv.bq4);
			break;
		default:
			debug2("GET skipped header %02x\n", hi);
			break;
		}
	}

	switch (handler("get")) {
	case -1:
		debug("GET handler error\n");
		exit(-1);
		break;
	case 1:
		debug("GET ok\n");
		if (context->data_length > 0) {
			// FIX ME (more header fields?)
			OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			hv.bs = context->data;
			OBEX_ObjectAddHeader(handle, oo, OBEX_HDR_BODY, hv, context->data_length, 0);
			hv.bq4 = context->data_length;
			OBEX_ObjectAddHeader(handle, oo, OBEX_HDR_LENGTH, hv, sizeof(uint32_t), 0);
		}
		else {
			debug("GET ok (body length <= 0)\n");
			OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
			hv.bq4 = context->data_length;
			OBEX_ObjectAddHeader(handle, oo, OBEX_HDR_LENGTH, hv, sizeof(uint32_t), 0);
			// FIX ME
		}
		break;
	default:
		debug("GET deny (fix me!)\n");
		// FIX ME
		OBEX_ObjectSetRsp(oo, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
		break;
	}

	context_clear(0);
	return(1);
	
err:
	debug("GET end - with error (fix me!)\n");
	// FIX ME (internal error)
	OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	context_clear(0);
	return(0);
}

int cmd_put(obex_t *handle, obex_object_t *oo)
{
	obex_headerdata_t hv;
	uint8_t hi;
	int hlen;


	debug("PUT start\n");

	while (OBEX_ObjectGetNextHeader(handle, oo, &hi, &hv, &hlen))	{
		switch (hi) {
		case OBEX_HDR_BODY:
			context->data = hv.bs;
			context->data_length = hlen;
			if (hlen > 0) context->data_type = 1;
			else context->data_type = 2;
			
			debug2("PUT body length: %d\n", hlen);
			break;
		case OBEX_HDR_NAME:
			if ((context->filename = malloc(hlen)))	{
				OBEX_UnicodeToChar(context->filename, hv.bs, hlen);

				debug2("PUT name: %s\n", context->filename);
			}
			else {
				// malloc error
				goto err;
			}
			break;
		case OBEX_HDR_TYPE:
			if ((context->type = malloc(hlen)))	{
				strncpy(context->type, hv.bs, hlen);
								
				debug2("PUT type: %s\n", context->type);
			}
			else {
				// malloc error
				goto err;
			}
			break;
		case OBEX_HDR_LENGTH:
			context->data_length = hv.bq4;
			
			debug2("PUT length: %d\n", context->data_length);
			break;
		case OBEX_HDR_TIME:
			debug2("PUT time: %u (fix me!)\n", hv.bq4);
			break;
		case OBEX_HDR_DESCRIPTION:
			if ((context->description = malloc(hlen)))	{
				OBEX_UnicodeToChar(context->description, hv.bs, hlen);
				
				debug2("PUT description: %s\n", context->description);
			}
			else {
				// malloc error
				goto err;
			}
			break;
		case OBEX_HDR_COUNT:
			context->count = hv.bq4;
			
			debug2("PUT count: %d\n", context->count);
			break;
		default:
			debug2("skipped header %02x\n", hi);
			break;
		}
	}

	if (context->filename == NULL)	{
		debug("PUT warning - no name!\n");
	}

	debug2("PUT data_type %d\n", context->data_type);

	switch(handler("put")) {
	case -1:
		debug("PUT handler error\n");
		exit(-1);
		break;
	case 1:
		OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
		debug("PUT ok\n");
		break;
	default:
		// FIX ME
		OBEX_ObjectSetRsp(oo, OBEX_RSP_FORBIDDEN, OBEX_RSP_FORBIDDEN);
		debug("PUT deny - fix me!\n");
		break;
	}

	debug("PUT end\n");
	context_clear(0);
	return(1);

err:
	debug("PUT end - with error (fix me!)\n");
	// FIX ME (internal error)
	OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
	context_clear(0);
	return(0);
}

void eventhandler(obex_t *handle, obex_object_t *oo, int mode, int event, int obex_cmd, int obex_rsp)
{
	int fd;
	unsigned int ssize;
	obex_internal_t *self = (obex_internal_t*) handle;
	struct sockaddr_rc tmp = {0};
	

	switch (event) {
	case OBEX_EV_PROGRESS:
		debug("PROGRESS\n");
		break;
	
	case OBEX_EV_ACCEPTHINT:
		debug("ACCEPTHINT\n");
		
		context_clear(1);
		
		fd = self->fd;
		ssize = sizeof(struct sockaddr_rc);
		getpeername(fd, (struct sockaddr*) &self->trans.peer.rfcomm, &ssize);
		bacpy(&context->remote_address, &self->trans.peer.rfcomm.rc_bdaddr);
		ssize = sizeof(struct sockaddr_rc);
		getsockname(fd, (struct sockaddr*) &tmp, &ssize);
		bacpy(&context->local_address, &tmp.rc_bdaddr);
		
		switch (handler("accept")) {
		case -1:
			debug("ACCEPTHINT handler error\n");
			exit(-1);
			break;
		case 1:
			debug("ACCEPTHINT ok\n");
			break;
		case 2:
			debug("ACCEPTHINT ok - require authentication\n");
			if (!cmd_auth(-1, &context->remote_address)) {
				debug("ACCEPTHINT can't authenticate peer\n");
				OBEX_TransportDisconnect(handle);
			}
			break;
		case 3:
			debug("ACCEPTHINT ok - require authentication and encryption\n");
			if (!cmd_auth(-1, &context->remote_address)) {
				debug("ACCEPTHINT can't authenticate peer\n");
				OBEX_TransportDisconnect(handle);
			}
			else {
				if (!cmd_enc(-1, &context->remote_address)) {
					debug("ACCEPTHINT can't encrypt connection\n");
					OBEX_TransportDisconnect(handle);
				}
			}
			break;
		default:
			debug("ACCEPTHINT deny\n");
			OBEX_TransportDisconnect(handle);
			break;
		}
		break;

	case OBEX_EV_REQDONE:
		switch (obex_cmd) {
		case OBEX_CMD_DISCONNECT:
			OBEX_TransportDisconnect(handle);
			break;
		default:
			debug("REQDONE\n");
			break;
		}
		break;
		
	case OBEX_EV_REQHINT:
		debug("REQHINT - add handler for this!\n");
		OBEX_ObjectSetRsp(oo, OBEX_RSP_CONTINUE, OBEX_RSP_SUCCESS);
		break;

	case OBEX_EV_REQ:
		switch (obex_cmd) {
		case OBEX_CMD_DISCONNECT:
			cmd_disconnect(handle, oo);
			break;
		case OBEX_CMD_CONNECT:
			cmd_connect(handle, oo);
			break;
		case OBEX_CMD_SETPATH:
			cmd_setpath(handle, oo);
			break;
		case OBEX_CMD_GET:
			cmd_get(handle, oo);
			break;
		case OBEX_CMD_PUT:
			cmd_put(handle, oo);
			break;
		default:
			debug("REQ unknown\n");
			break;
		}
		break;
		
	case OBEX_EV_LINKERR:
		debug("LINKERR\n");
		break;
	
	case OBEX_EV_UNEXPECTED:
		// comes together with EV_PROGRESS ...
		break;
	
	default:
		printf("unknown event %d\n", event);
		break;
	}
}

void usage(char *progname)
{
	fprintf(stderr, "syntax: %s -achuofdpvH\n"\
		"-a bdaddr   local address\n"\
		"-c channel  rfcomm channel (default is first available)\n"\
		"-h handler  path to handler executable\n"\
		"-u user     user to run handler as (default is nobody)\n"\
		"-o          don't register OPUSH\n"\
		"-f          don't register OBEXFTP\n"\
		"-d          debug (don't become a daemon)\n"\
		"-p          print used rfcomm channel\n"\
		"-v          print version\n"\
		"-H          print this help message\n"\
		"", progname);
}

int main(int argc, char **argv)
{
	obex_t *libObex;
	int result;
	int opt;
	int print_channel = 0;
	
	
	context = (server_context*) malloc(sizeof(server_context));
	memset(context, 0, sizeof(server_context));
	context->sdp_opush = 1;
	context->sdp_ftp = 1;
	context->daemon = 1;

	while ((opt = getopt(argc, argv, "a:c:h:u:ofdpvH")) != EOF) {
		switch (opt) {
		case 'a':
			if (!str2ba(optarg, &context->local_address)) {
				fprintf(stderr, "%s is not a valid address\n", optarg);
				exit(-1);
			}
			break;
		case 'c':
			context->channel = atoi(optarg);
			break;
		case 'h':
			context->handler_script = optarg;
			break;
		case 'o':
			context->sdp_opush = 0;
			break;
		case 'f':
			context->sdp_ftp = 0;
			break;
		case 'd':
			context->daemon = 0;
			break;
		case 'p':
			print_channel = 1;
			break;
		case 'v':
			fprintf(stderr, VERSION_STRING);
			exit(1);
			break;
		case 'u':
			context->user = getpwnam(optarg);
			if (context->user == NULL) {
				fprintf(stderr, "%s is not a valid user\n", optarg);
				usage(argv[0]);
				exit(-1);
			}
			break;
		case 'H':
			usage(argv[0]);
			exit(1);
		default:
			usage(argv[0]);
			exit(-1);
		}
	}
	
	if (optind < argc) {
		usage(argv[0]);
		exit(-1);
	}
	
	if (context->handler_script == NULL) {
		context->handler_script = HANDLER;
	}

	if (context->user == NULL) {
		context->user = getpwnam("nobody");
		if (context->user == NULL) {
			fprintf(stderr, "user nobody doesn't exists\n");
			usage(argv[0]);
			exit(-1);
		}
	}

	if (access(context->handler_script, X_OK) != 0) {
		fprintf(stderr, "can't use given handler: %s\n", context->handler_script);
		usage(argv[0]);
		exit(-1);
	}

	libObex = OBEX_Init(OBEX_TRANS_BLUETOOTH, eventhandler, OBEX_FL_KEEPSERVER);
	if (libObex == NULL) {
		fprintf(stderr, "can't initialize obex\n");
		exit(-1);
	}
	
	OBEX_SetCustomData(libObex, (void*)context);
	
	result = BtOBEX_ServerRegister(libObex, &context->local_address, context->channel);
	if (!result) {
		fprintf(stderr, "can't bind to address or channel\n");
		usage(argv[0]);
		exit(-1);
	}
	
	/* get channel for dynamic channel allocation */
	if (context->channel == 0) {
		struct sockaddr_rc sa = {0};
		int sa_len = sizeof(struct sockaddr_rc);
		getsockname(OBEX_GetFD(libObex), (struct sockaddr*)&sa, &sa_len);
		context->channel = sa.rc_channel;
	}
	
	if (print_channel) printf("%d\n", context->channel);
	
	if (context->sdp_opush) add_opush();
	if (context->sdp_ftp) add_ftp();
	
	/* become a daemon */
	if (context->daemon) {
		int fd;
		if (fork()) exit(0);
		fd = open("/dev/null", O_RDWR);
		dup2(fd, 0);
		dup2(fd, 1);
		dup2(fd, 2);
		close(fd);
		setsid();
		chdir("/");
	}
	
	while (OBEX_HandleInput(libObex, 1) != -1);

	return(1);
}

/*
 *  $Log$
 */
