/*
 *  $Id$
 */

/*
 *  sobexsrv (c) Collin R. Mulliner <collin-sobexsrv(AT)betaversion.net>
 *               web: http://www.mulliner.org/bluetooth/
 *
 *  license: GPL
 *
 */

#include <bluetooth/bluetooth.h>
#include <bluetooth/rfcomm.h>
#include <netinet/in.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#ifndef __OBEXSRV_H__
#define __OBEXSRV_H__

/* start OPENOBEX structs - DONT CHANGE! */
typedef union
{
	struct sockaddr_in inet;
	struct sockaddr_rc rfcomm;
} saddr_t;

typedef struct obex_transport
{
	int type;
	int connected;    /* Link connection state */
	unsigned int mtu; /* Tx MTU of the link */
	saddr_t self;     /* Source address */
	saddr_t peer;     /* Destination address */
} obex_transport_t;

typedef struct obex_internal
{
	uint16_t mtu_tx;     /* Maximum OBEX TX packet size */
	uint16_t mtu_rx;     /* Maximum OBEX RX packet size */
	uint16_t mtu_tx_max; /* Maximum TX we can accept */

	int fd;             /* Socket descriptor */
	int serverfd;
	int writefd;        /* write descriptor - only OBEX_TRANS_FD */
	unsigned int state;
        
	int keepserver; /* Keep server alive */
	int filterhint; /* Filter devices based on hint bits */
	int filterias;  /* Filter devices based on IAS entry */

	void *tx_msg; /* Reusable transmit message */
	void *rx_msg; /* Reusable receive message */

	obex_object_t *object; /* Current object being transfered */      
	obex_event_t eventcb;  /* Event-callback */

	obex_transport_t trans; /* Transport being used */
	obex_ctrans_t ctrans;
	void *userdata;         /* For user */
} obex_internal_t;
/* end OPENOBEX structs - DONT CHANGE! */

#define UNNAMED_PUT_FILENAME "unnamed_file"

typedef struct server_context_t
{
	char *handler_script;
	int mode;             /**< 0=opush 1=ftp */
	struct passwd *user;
	
	bdaddr_t local_address;       /**< BD address to bind */
	unsigned short int channel;   /**< the RFCOMM channel */
	unsigned short int sdp_opush; /**< publish OPUSH record in SDP */
	unsigned short int sdp_ftp;   /**< publish OBEXFTP record in SDP */
	
	sdp_record_t *sdp_record[2]; /**< SDP record */
	sdp_session_t *sdp_session;  /**< SDP session */

	short int keep_running; /**< don't exit after handling first client */

	unsigned int connection_id; /**< connection id */
	bdaddr_t remote_address;
	
	char *target;      /**< target */
	int target_length; /**< length of target */
	
	char *filename;    /**< filename get/put */
	char *path;        /**< path (setpath) */ 
	char *type;        /**< type */
	char *description; /**< description */
	int count;         /**< count */

	uint8_t *data;            /**< data */
	int data_type;            /**< type of data 0=nodata, 1=data, 2=empty*/
	unsigned int data_length; /**< length of data */
	
	int daemon;
} server_context;

#endif

/*
 *  $Log$
 */
